#ifndef lint
static char SCCSid[] = "@(#) ./comm/io/pibase.c 12/01/93";
#endif

/*
   This file contains simple routines for writing and reading sequentially
   to a file.  If the file is distributed, sequentially applies to each task.
   If the file is not distributed, then the elements are read/written in task 
   rank order (that is, the read/write operations here are COLLECTIVE; all
   tasks in the procset must call them together.

   In the AS_SEQUENTIAL case, the algorithm is

   if (PSISROOT()) {
       write my contribution
       post receive
       for (i=0; i<np; i++) {
           PIgtoken(pset,i)
	   wait for data
	   post next receive
	   write data
	   }
	}
    else {
        for (i=0; i<np; i++) {
	    if (PIgtoken(pset,i)) send data to master
	    }
	}
 */

#include <stdio.h>
#include <ctype.h>
#if defined(cray)
#include <errno.h>
#else
extern int errno;
#endif
#include <sys/errno.h>
#include "tools.h"
#include "comm/comm.h"
#include "comm/io/pio.h"
#ifndef SEEK_SET
#include <unistd.h>
#endif

#define READ  (*fp->read)
#define WRITE (*fp->write)
#define LSEEK (*fp->lseek)

/*@
    PIFwrite - Write a block of data to a file.

    Input Parameters:
.   fp - Parallel file pointer
.   buffer - data to write
.   len    - length of data (in bytes)
.   datatype - data type (MSG_INT etc).  

    Notes:
    This is a collective routine.  All tasks must call it.
    In the case of an PIO_AS_SEQUENTIAL file, the blocks are written in 
    rank order (that is, for the process set PSAllNodes, first task 0
    writes, then task 1 writes, etc).  In the PIO_AS_PARALLEL mode, 
    many tasks MAY write simultaneously.
@*/
void PIFwrite( fp, buffer, len, datatype )
PIFILE *fp;
char   *buffer;
int    len, datatype;
{
char *p;
int  lenin, i, err, np;

np = PSnumtids(fp->procset);

if (fp->pmode == PIO_AS_SEQUENTIAL) {
    if (PSISROOT(fp->procset)) {
	err = WRITE( fp, buffer, len );
	if (err != len) {
	    SETERRC(1,"Write failed");
	    }
	PIgtoken(fp->procset,0);
	for (i=1; i<=np; i++) {
	    PIgtoken(fp->procset,i);
	    if (i < np) {
		/* We are ALWAYS active */
		PIbrecvUnsz( fp->mtype, p, lenin, datatype );
		err = WRITE( fp, p, lenin );
		if (err != lenin) {
		    SETERRC(1,"Write failed");
		    }
		PIFreeRecvBuf( p );
		}
	    }
	}
    else {
        for (i=0; i<=np; i++) {
	    if (PIgtoken(fp->procset,i)) 
		PIbsend( fp->mtype, buffer, len, PSROOT(fp->procset), 
		 	 datatype );
	    }
	}
    }
else {
    /* AS_PARALLEL, everyone writes directly */
    err = WRITE( fp, buffer, len );
    if (err != len) {
	SETERRC(1,"Write failed");
	}
    }
	
}

/*@
    PIFread - Read a block of data from a file.

    Input Parameters:
.   fp - Parallel file pointer
.   buffer - for data read
.   len    - length of data to read (in bytes)
.   datatype - data type (MSG_INT etc).    

    Notes:
    This is a collective routine.  All tasks must call it.
    In the case of an PIO_AS_SEQUENTIAL file, the blocks are read in 
    rank order (that is, for the process set PSAllNodes, first task 0
    reads, then task 1 reads, etc).  In the PIO_AS_PARALLEL mode, 
    many tasks MAY read simultaneously.
@*/
int PIFread( fp, buffer, len, datatype )
PIFILE *fp;
char   *buffer;
int    len, datatype;
{
char *p;
int  lenin, i, curlen, err, np;

np = PSnumtids(fp->procset);
if (fp->pmode == PIO_AS_SEQUENTIAL) {
    if (PSISROOT(fp->procset)) {
	err = READ( fp, buffer, len );
	if (err != len) {
	    SETERRC(1,"Read failed");
	    }
	PIgtoken(fp->procset,0);
	curlen = -1;
	p      = 0;
	for (i=1; i<=np; i++) {
	    PIgtoken(fp->procset,i);
	    if (i < np) {
		/* We are ALWAYS active */
		PIbrecv( fp->mtype, &lenin, sizeof(int), MSG_INT );
		if (curlen < lenin) {
		    if (p) FREE(p);
		    p      = MALLOC( lenin );   CHKPTR(p);
		    curlen = lenin;
		    }
		err = READ( fp, p, lenin );
		if (err != lenin) {
		    SETERRC(1,"Read failed");
		    }
		PIbsend( fp->mtype, p, lenin, PIfrom(), datatype );
		}
	    FREE( p );
	    }
	}
    else {
        for (i=0; i<=np; i++) {
	    if (PIgtoken(fp->procset,i)) {
		PIbsend( fp->mtype, &len, sizeof(int), PSROOT(fp->procset), 
		 	 MSG_INT );
		PIbrecv( fp->mtype, buffer, len, datatype );
		}
	    }
	}
    }
else {
    /* AS_PARALLEL, everyone reads directly */
    err = READ( fp, buffer, len );
    if (err != len) {
	SETERRC(1,"Read failed");
	}
    }

}
